/*****************************************************************
 *                                                               *
 *        AY  YM          *
 *                                                               *
 *  2002-2003   <sashnov@ngs.ru>                  *
 *  http://sashnov.chat.ru                                       *
 *                                                               *
 *    AY (  4.0 beta 0               *
 *  (c) 1998-2003 Roman Scherbakov AKA V_Soft                    *
 *  http://vtx.microfor.ru/                                      *
 *                                                               *
 *    AY  YM                                    *
 *  (c) by V_Soft and Lion 17 || (c) Hacker KAY                  *
 *                                                               *
 *                                          *
 *  (c) Hacker KAY & Sergey Bulba                                *
 *                                                               *
 *****************************************************************/

#include "ay8912.h"

typedef struct
{
        int     OutputFreq;             //  
        int     OutputChannels;         //   (1 | 2)
        int     OutputBPC;              // bps (8 | 16)
        int     ChipFreq;               //  
        chip_t  ChipType;               //   (AY | YM)
        stereo_t Stereo;                // 
} VCONFIG;

typedef struct
{
        Word    BitA;
        Word    BitB;
        Word    BitC;
        Word    BitN;
        Word    Bit0;
        Word    Bit1;
        Word    Bit2;
        Word    Bit3;
        Word    Bit4;
        Word    Bit5;
} BITSTR;

typedef struct
{
        int     Global;                 //   
        int     A_l;
        int     A_r;
        int     B_l;
        int     B_r;
        int     C_l;
        int     C_r;
} AMPC;

typedef struct
{
        int     CntA;
        int     CntB;
        int     CntC;
        int     CntE;
        int     CntN;
} COUNT;

static int bEnvGenInit = 0;		/*    */
static int bEQSet = 0;			/*    */

static  void GenVol(void);              //   
static  void GenEnv(void);              //  



/*   ===   AY === */

static  AYREGS  Ay;                     // 
static  COUNT   Counter;                //    AY
static  int     EnvNum;                 //    ( 0  15)
static  int     EnvPos;                 //      ( 0  127)
static  int     Cur_Seed=0xffff;        //     



/*    
  (  44100 , 16 , ,  AY,   ABC) */
static  VCONFIG config = {44100, 2, 16, 1773400, _AY, _ABC};

//   
static  BITSTR  bitstream;

//   
static  AMPC    Amp;



/*  ===  (  )  ===  */

/*    ,      , . AyReset() */
static  int     ChipTacts_per_outcount;

/*     ,   GenVol()
          
          ( AmpA, AmpB  AmpC )
       32 , . AyRCorr()
 */
static  int     vols[6][32];

/*   .
      GenEnv()
     128   , . AyRCorr()
 */
static  int     Envelope[16][128];

/*   ,     */
static  Word AYVol[32], YMVol[32];






/*       AY  YM
    (  AY 16 ,  YM 32 ) */
//
// (c) by V_Soft and Lion 17
//
static  Word AYVol_L17[16] = // AY Volume
        {0, 513, 828, 1239, 1923, 3238, 4926, 9110, 10344, 17876, 24682, 30442,
        38844, 47270, 56402, 65535};

static  Word YMVol_L17[32] = // YM Volume 
        {0,  0,  190,  286, 375, 470,  560, 664, 866, 1130, 1515, 1803, 2253,
        2848, 3351, 3862, 4844, 6058, 7290, 8559, 10474, 12878, 15297, 17787,
        21500, 26172, 30866, 35676, 42664, 50986, 58842, 65535};

//
// (c) by Hacker KAY
//
static  Word AYVol_KAY[16] = // AY Volume
        {0, 836, 1212, 1773, 2619, 3875, 5397, 8823, 10392, 16706, 23339, 29292,
        36969, 46421, 55195, 65535};

static  Word YMVol_KAY[32] = // YM Volume 
        {0, 0, 248, 450, 670, 826, 1010, 1239, 1552, 1919, 2314, 2626, 3131, 3778,
        4407, 5031, 5968, 7161, 8415, 9622, 11421, 13689, 15957, 18280, 21759,
        26148, 30523, 34879, 41434, 49404, 57492, 65535};



static  int Layout[2][7][6] =
{
//     AY chip
//       A_l  A_r  B_l  B_r  C_l  C_r
       {{100, 100, 100, 100, 100, 100},         // _MONO
        {100,  33,  70,  70,  33, 100},         // _ABC
        {100,  33,  33, 100,  70,  70},         // _ACB
        { 70,  70, 100,  33,  33, 100},         // _BAC
        { 33, 100, 100,  33,  70,  70},         // _BCA
        { 70,  70,  33, 100, 100,  33},         // _CAB
        { 33, 100,  70,  70, 100,  33}},        // _CBA
//     YM chip
//       A_l  A_r  B_l  B_r  C_l  C_r
       {{100, 100, 100, 100, 100, 100},         // _MONO
        {100,   5,  70,  70,   5, 100},         // _ABC
        {100,   5,   5, 100,  70,  70},         // _ACB
        { 70,  70, 100,   5,   5, 100},         // _BAC
        {  5, 100, 100,   5,  70,  70},         // _BCA
        { 70,  70,   5, 100, 100,   5},         // _CAB
        {  5, 100,  70,  70, 100,   5}}         // _CBA
};


/*     */
int AY_SetSoundFormat(int freq,int chans, int bpc)
{
        if ((bpc != 16 && bpc != 8) || (chans != 2 && chans != 1))
                return 0;
        
        config.OutputFreq = freq;
        config.OutputChannels = chans;
        config.OutputBPC = bpc;

        return 1;
}

//      
//       [-100..100],  
//          
void AY_SetEQ(int Amp_A_l, int Amp_A_r, int Amp_B_l, int Amp_B_r, int Amp_C_l, int Amp_C_r)
{
        int def_Amp_A_l, def_Amp_A_r;
        int def_Amp_B_l, def_Amp_B_r;
        int def_Amp_C_l, def_Amp_C_r;
        int n;

        n = (config.ChipType == _AY) ? 0 : 1;

        def_Amp_A_l = Layout[n][config.Stereo][0];
        def_Amp_A_r = Layout[n][config.Stereo][1];
        def_Amp_B_l = Layout[n][config.Stereo][2];
        def_Amp_B_r = Layout[n][config.Stereo][3];
        def_Amp_C_l = Layout[n][config.Stereo][4];
        def_Amp_C_r = Layout[n][config.Stereo][5];

        Amp.A_l = ((Amp_A_l < -100) || (Amp_A_l > 100)) ? def_Amp_A_l : Amp_A_l;
        Amp.A_r = ((Amp_A_r < -100) || (Amp_A_r > 100)) ? def_Amp_A_r : Amp_A_r;
        Amp.B_l = ((Amp_B_l < -100) || (Amp_B_l > 100)) ? def_Amp_B_l : Amp_B_l;
        Amp.B_r = ((Amp_B_r < -100) || (Amp_B_r > 100)) ? def_Amp_B_r : Amp_B_r;
        Amp.C_l = ((Amp_C_l < -100) || (Amp_C_l > 100)) ? def_Amp_C_l : Amp_C_l;
        Amp.C_r = ((Amp_C_r < -100) || (Amp_C_r > 100)) ? def_Amp_C_r : Amp_C_r;

        bEQSet = 1;
}

//   ,    ,
//  -1    .
void AY_SetChip(chip_t chip, int chipfreq, stereo_t stereo)
{
        config.ChipType = (chip     == -1)? _AY     : chip;
        config.Stereo   = (stereo   == -1)? _ABC    : stereo;
        config.ChipFreq = (chipfreq == -1)? 1773400 : chipfreq;
}

//    
//    AY 16 
void AY_SetAYTable(Word *tbl)
{
        int n;

        /* ..        32 
             */
        for(n=0; n<16; n++) {
                AYVol[n*2+1] = AYVol[n*2] = tbl[n];
        }
}

//     
//    YM 32 
void AY_SetYMTable(Word *tbl)
{
        int n;

        for(n=0; n<32; n++)
                AYVol[n] = tbl[n];
}

//  C ,     
void AY_Start(void)
{
        int vol, max_l, max_r;

        //        
        if (!bEnvGenInit)
                GenEnv();       

        //    ,   
        if (!bEQSet)
                AY_SetEQ(255, 255, 255, 255, 255, 255); //       
        
        //   skipped ;-)
        ChipTacts_per_outcount = config.ChipFreq / config.OutputFreq / 8;

        //     
        Counter.CntA = Counter.CntB = Counter.CntC = Counter.CntN = Counter.CntE = 0;
        Cur_Seed = 0xFFFF;

        //   
        bitstream.BitA = 0;
        bitstream.BitB = 0;
        bitstream.BitC = 0;
        bitstream.BitN = 0;

        //  
        EnvNum = 0;
        EnvPos = 0;

        /*      .
               
            AY  YM   
          */ 

        //     
        AY_SetAYTable((Word *)&AYVol_KAY);
        AY_SetYMTable((Word *)&YMVol_KAY);

        GenVol();

        //     
        // ,   vols [x][31]    
        // TODO:     ;-)
        max_l = vols [0][31] + vols [2][31] + vols [3][31];
        max_r = vols [1][31] + vols [3][31] + vols [5][31];
        vol = (max_l > max_r) ? max_l : max_r;
        Amp.Global = ChipTacts_per_outcount * vol / MAX_AMP;
}

//        
void AY_SetRegs(char *regs)
{
        int n;
        
        //  
        for( n=0; n<14; n++)
                Ay.Regs[n] = regs [n];
        
        //      4095
        Ay.Values.toneA &= 4095;
        Ay.Values.toneB &= 4095;
        Ay.Values.toneC &= 4095;

        //      31
        Ay.Values.noise &= 31;

        //         
        // ..     ,   word- (0  65535)
        bitstream.Bit0 = 0 - ((Ay.Values.control>>0) & 1);
        bitstream.Bit1 = 0 - ((Ay.Values.control>>1) & 1);
        bitstream.Bit2 = 0 - ((Ay.Values.control>>2) & 1);
        bitstream.Bit3 = 0 - ((Ay.Values.control>>3) & 1);
        bitstream.Bit4 = 0 - ((Ay.Values.control>>4) & 1);
        bitstream.Bit5 = 0 - ((Ay.Values.control>>5) & 1);

        //   
        /*   0...3     0  15.
         4 ( 16)  ,  . */
        
        Ay.Values.volumeA &= 31;
        Ay.Values.volumeB &= 31;
        Ay.Values.volumeC &= 31;
        
        // 255 -   
        //    ,
        if( Ay.Values.env_style != 255 )
        {
                EnvNum = Ay.Values.env_style & 15;
                EnvPos = 0;
                Counter.CntE = 0;
        }
}

/*   numcount   
(     16-  8-  
      ) */
Byte* AY_Synth(Byte *outbuf_pos, int numcount)
{
        int     mix_l, mix_r;           //  
        int     vol;
        int     vol_e;
        Word    bit;
        int     n;
        Byte *outbuf_b;
	Word *outbuf_w;
        
        do {
                mix_l = mix_r = 0;
                outbuf_b = outbuf_pos;
        
                //   AY     
                for( n=0; n < ChipTacts_per_outcount; n++ )
                {
                        // Gen tones
                        if( ++Counter.CntA >= Ay.Values.toneA ) bitstream.BitA ^= -1, Counter.CntA = 0;
                        if( ++Counter.CntB >= Ay.Values.toneB ) bitstream.BitB ^= -1, Counter.CntB = 0;
                        if( ++Counter.CntC >= Ay.Values.toneC ) bitstream.BitC ^= -1, Counter.CntC = 0;

                        // GenNoise
                        // (c) Hacker KAY & Sergey Bulba
                        if( ++Counter.CntN >= (Ay.Values.noise << 1))
                        {
                                Cur_Seed = (Cur_Seed * 2 + 1) ^ (((Cur_Seed >> 16) ^ (Cur_Seed >> 13)) & 1);
                                bitstream.BitN = 0 - ((Cur_Seed >> 16) & 1);
                                Counter.CntN = 0;
                        }
                        
                        // Gen Envelope
                        if( ++Counter.CntE >= Ay.Values.env_freq )
                        {
                                if( ++EnvPos >= 128 )
                                        EnvPos = 64;
                                Counter.CntE = 0;
                        }

                        vol_e = Envelope[ EnvNum ][ EnvPos ];

                        // Mixer (NEW)
                        vol = (Ay.Values.volumeA & 0x10) ? vol_e : (Ay.Values.volumeA & 0x0F) * 2 + 1;
                        bit = (bitstream.BitA | bitstream.Bit0) & (bitstream.BitN | bitstream.Bit3);
                        if (bit) {
                                mix_l += vols[0][vol];
                                mix_r += vols[1][vol];
                        }

                        vol = (Ay.Values.volumeB & 0x10) ? vol_e : (Ay.Values.volumeB & 0x0F) * 2 + 1;
                        bit = (bitstream.BitB | bitstream.Bit1) & (bitstream.BitN | bitstream.Bit4);
                        if (bit) {
                                mix_l += vols[2][vol];
                                mix_r += vols[3][vol];
                        }

                        vol = (Ay.Values.volumeC & 0x10) ? vol_e : (Ay.Values.volumeC & 0x0F) * 2 + 1;
                        bit = (bitstream.BitC | bitstream.Bit2) & (bitstream.BitN | bitstream.Bit5);
                        if (bit) {
                                mix_l += vols[4][vol];
                                mix_r += vols[5][vol];
                        }

                }
                
                // Interpol
                mix_l /= Amp.Global;
                mix_r /= Amp.Global;

                //     
                if ( config.OutputBPC == 8)
                {
                        mix_l = (mix_l >> 8) | 128;
                        mix_r = (mix_r >> 8) | 128;

                        *(outbuf_b)++ = mix_l;
                        if ( config.Stereo != _MONO )
                                *(outbuf_b)++ = mix_r;

                        *outbuf_pos = *outbuf_b;  // ???
                }
                else
                {
                        outbuf_w=outbuf_pos;
                        *(outbuf_w)++ = mix_l;
                        if ( config.Stereo != _MONO ) {
                                *(outbuf_w)++ = mix_r;
				  }
                        outbuf_pos=outbuf_w;
                }
        } 
        while (--numcount);

        return outbuf_pos;
}

//          
//       
static void GenVol()
{
        int n;
        int vol;

        for( n=0; n<32; n++ )
        {
                vol = ( config.ChipType == _AY ) ? AYVol[n] : YMVol[n];

                vols [0][n] = (int)(((double)vol * Amp.A_l) / 100);
                vols [1][n] = (int)(((double)vol * Amp.A_r) / 100);
                vols [2][n] = (int)(((double)vol * Amp.B_l) / 100);
                vols [3][n] = (int)(((double)vol * Amp.B_r) / 100);
                vols [4][n] = (int)(((double)vol * Amp.C_l) / 100);
                vols [5][n] = (int)(((double)vol * Amp.C_r) / 100);
        }
}

//  
static void GenEnv()
{
        int env, pos;
        int Hold;
        int Dir;
        int Vol;

        bEnvGenInit = 1;

        //    
        for( env = 0; env<16; env++)
        {
                Hold = 0; //  
                //    ? ( 2)
                Dir = (env & 4)?  1 : -1;
                Vol = (env & 4)? -1 : 32;
                //  
                for( pos=0; pos<128; pos++)
                {
                        //   
                        if( !Hold )
                        {
                                //       
                                Vol += Dir;
                                //      
                                if ( Vol < 0 || Vol >= 32 )
                                {
                                        //    3 - 
                                        if ( env & 8 )
                                        {
                                                //     2- 
                                                //  
                                                if ( env & 2 ) Dir = -Dir;
                                                Vol = ( Dir > 0 ) ? 0:31;
                                                //    0-
                                                //  
                                                if ( env & 1 )
                                                {
                                                        Hold = 1;
                                                        Vol = ( Dir > 0 ) ? 31:0;
                                                }
                                        }
                                        // ,    , 
                                        //   
                                        else
                                        {
                                                Vol = 0;
                                                Hold = 1;
                                        }
                                }
                        }
                        Envelope[env][pos] = Vol;
                }
        }
}
