/*
      VTX
       
*/
#include "vtxfile.h"
#include <string.h>
#include <stdlib.h>

#if 1
#include <glib.h>
#define malloc(x) g_malloc(x)
#define realloc(x,y) g_realloc(x,y)
#define free(x) g_free(x)
#endif


//    lh5dec.c
extern void lh5_decode(unsigned char *inp,unsigned char *outp,long original_size,long packed_size);


/*  NTString   
  NTString -   0  254 , . 0 */
static void getntstring(char* ntstring, FILE* fp)
{
    char c;
    int count=0;

	do{
		c = fgetc(fp);
		if (++count == VTX_NTSTRING_MAX)
			c = 0;
		*ntstring++ = c;
	} while (c != 0);
}


//       VTX
//       
//    VthHeader.
//     true (1)
//    false (0)
int VTX_ReadHeader(VTXFile *vf, char *filename)
{
    char buf[2];

    vf->regdata = NULL;
    
    if ((vf->fp = fopen(filename,"rb")) == NULL) {
		fprintf(stderr,"Cannot open file %s\n",filename);
		return 0;
    }

    //      AY  YM (.  VTX)
    fread(buf,2,1,vf->fp);
	if (strncmp(buf,"ay",2) == 0 || strncmp(buf,"AY",2))
		vf->Chip = "ay";
	else if (strncmp(buf,"ym",2) == 0 || strncmp(buf,"YM",2))
		vf->Chip = "ym";
	else {
		fprintf(stderr,"File %s is _not_ VORTEX format!\nIt not begins from AY or YM.\n",filename);
		fclose(vf->fp);
		vf->fp = NULL;
		return 0;
    }
	
    //  1    0-7 (MONO, ABC,....)
    vf->Stereo = fgetc(vf->fp);

    //     (   )
    fread( &vf->Loop, 2, 1, vf->fp);
    
    //   
    fread( &vf->ChipFreq, 4, 1, vf->fp);
    
    //  
    fread( &vf->PlayerFreq, 1, 1, vf->fp);
        
    //   
    fread( &vf->Year, 2, 1, vf->fp);

    //   
    fread( &vf->regdata_size, 4, 1, vf->fp);

    // NTString 
    getntstring(vf->Title,	vf->fp);	// 
    getntstring(vf->Author,	vf->fp);	// 
    getntstring(vf->From,	vf->fp);	// 
    getntstring(vf->Tracker,vf->fp);	// 
    getntstring(vf->Comment,vf->fp);	// 

    return 1;
}



/*     .
	  ,    
	VTX_ReadHeader  . 
	 1  , 0    */
int VTX_ReadData(VTXFile *vf)
{
    //       
    Byte * packed_data;
    int packed_size, maxbuf;

	//    
	if (vf->fp == NULL)
		return 0;
    	
    packed_size=0;
    maxbuf = 1;
    packed_data = (Byte*)malloc(maxbuf);

	//      
    while (!feof(vf->fp))
    {
		if (++packed_size > maxbuf) {
			maxbuf*= 2;
			packed_data = (Byte*)realloc(packed_data,maxbuf);
			if (packed_data == NULL) {
				fprintf(stderr,"VTX_ReadData: Packed data out ofmemory!\n");
				fclose(vf->fp);
				return 0;
			}
		}
		//   	
		packed_data[packed_size-1] = fgetc(vf->fp);
    };

    //  
    fclose(vf->fp);
    vf->fp = NULL;

    //     
    if ((vf->regdata = (Byte*)malloc(vf->regdata_size)) == NULL) {
		fprintf(stderr,"VTX_ReadData: No memory for regdata!\n");
		free(packed_data);
		return 0;
	}

    lh5_decode( packed_data /* in */ , vf->regdata /* out */, 
		vf->regdata_size /* inputlen */, packed_size /*outputlen */ );

    free(packed_data);

    //    
    vf->pos = 0;	//      

    return 1;
}


/*      
  1,    
 0, ,    */
int VTX_GetNextRegs(VTXFile *vf,char *regs)
{
	int n,reglen;
	Byte *p;

	reglen = vf->regdata_size/14;

	if (vf->pos >= reglen)
 
		return 0;

	p = &vf->regdata[vf->pos];
	for(n=0; n<14; n++, p += reglen)
		regs[n] = *p;
	vf->pos++;
	return 1;
}



/*      
	 VTXFile,     */
void VTX_Free(VTXFile *vf)
{
	if (vf->fp)
 {
		fclose(vf->fp);
		vf->fp = NULL;
	}
	
	if (vf->regdata)
 {
		free(vf->regdata);
		vf->regdata = NULL;
	}
}
