/*  VTXplugin - VTX player for XMMS
 *
 *  Copyright (C) 2002-2003 Sashnov Alexander
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/util.h>

#include "xmms-vtx.h"
#include "vtxfile.h"
#include "ay8912.h"

#define OUT_BUFSIZE 4096

//  
gboolean going = FALSE;
gboolean audio_error = FALSE;
pthread_t play_thread;

int seek_to;			/*     */

/* ,      */
const int OUT_FREQ = 44100;
const int BYTES_PER_SAMPLE = 4;	// 16 bit stereo


static Byte OutBuffer[OUT_BUFSIZE];	/*   */

/*  ===     === */

VTXFile vtx_file;		/* current play tune */
int samples_per_regs;		/*      */
int end;			/*    */



// For export
static InputPlugin vtx_ip = {
  NULL,
  NULL,
  NULL,				/* Description */
  NULL,				// vtx_init,
  vtx_about,
  NULL,
  vtx_is_our_file,
  NULL,
  vtx_play_file,
  vtx_stop,
  vtx_pause,
  vtx_seek,
  NULL,
  vtx_get_time,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  vtx_get_song_info,
  NULL,				/* file_info_box */
  NULL
};

// called from xmms for plug
InputPlugin *
get_iplugin_info (void)
{
  vtx_ip.description = g_strdup_printf (_("Vortex file Player %s"), VERSION);
  return &vtx_ip;
}

void
vtx_about (void)
{
  static GtkWidget *box;
  box = xmms_show_message (_("About Vortex Player"),
			   _
			   ("Vortex Player by Sashnov Alexander <sashnov@ngs.ru>\n"
			    "Original source in_vtx.dll from Roman Sherbakov <>\n\n"
			    "Music in vtx format can be found at v_soft.chat.ru\n"
			    "and other sites."), _("Ok"), FALSE, NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (box), "destroy", gtk_widget_destroyed,
		      &box);
}


/* Return 1 if the plugin can handle the file */
int
vtx_is_our_file (char *filename)
{
  char *p;
  char buf[2];
  FILE *fp;

  // search end of string filename
  for (p = filename; *p != 0; p++);

  // back to 4 from end symbol
  p -= 4;

  // if file extension not ".vtx" return FALSE
  if (strcmp (p, ".vtx"))
    return 0;

// DEBUG: for faster, not read file
  return 1;

  // filename OK, open this file
  fp = fopen (filename, "rb");

  // read first 2 bytes from file
  fread (buf, 2, 1, fp);

  // close file
  fclose (fp);

  // if file beginig from "ay" or "ym" ignoring case than this is vtx format file
  return (!strncasecmp (buf, "ay", 2) || !strncasecmp (buf, "ym", 2));
}


/*  ,   vtx_play_file() */
void *
play_loop (void *args)
{
  Byte *stream;			/*       */
  unsigned char regs[14];	/*     */
  int need;			/*       */
  int left = 0;			/*            */
  int donow;



#if 1

  /*    */
  while (going)
    {

      if (end)
	xmms_usleep (10000);
      else
	{
	  /*        need       */
	  stream = &OutBuffer[0];
	  need = OUT_BUFSIZE / BYTES_PER_SAMPLE;

	  /*       */
	  while (need > 0)
	    {
	      if (left == 0)	/*       */
		{
		  if (VTX_GetNextRegs (&vtx_file, regs) != 0)
		    {
		      left = samples_per_regs;
		      AY_SetRegs (regs);
		    }
		  else
		    {
		      //   ,     
		      //       .
		      end = 1;
		      memset (stream, need, BYTES_PER_SAMPLE);
		      break;	/*      */
		    }
		}

	      //            
	      donow = (need > left) ? left : need;
	      stream = AY_Synth ((Byte*)stream, donow);
	      left -= donow;
	      need -= donow;
	    }			/* end of while(need > 0) */

	  /*       */
	  vtx_ip.add_vis_pcm (vtx_ip.output->written_time (), FMT_S16_NE,
			      2 /*channels */ , OUT_BUFSIZE, OutBuffer);

	  /*         */
	  while (vtx_ip.output->buffer_free () < OUT_BUFSIZE && going
		 && seek_to == -1)
	    xmms_usleep (10000);

	  if (going && seek_to == -1)
	    vtx_ip.output->write_audio (OutBuffer, OUT_BUFSIZE);

	  /*      ,    */
	  if (end)
	    {
	      vtx_ip.output->buffer_free ();
	      vtx_ip.output->buffer_free ();
	    }

	}

      /* jump to time in seek_to (in seconds) */
      if (seek_to != -1)
	{
	  vtx_file.pos = seek_to * 50;	/* time in sec * 50 = AY reg data blocks */
	  vtx_ip.output->flush (seek_to * 1000);
	  seek_to = -1;
	}
    }				/* end of while(going) */






#else


  while (going)
    {
      if (!end)
	{

	  /*        need       */
	  stream = &OutBuffer[0];
	  need = OUT_BUFSIZE / BYTES_PER_SAMPLE;

	  /*       */
	  while (need > 0)
	    {
	      if (left == 0)	/*       */
		{
		  if (VTX_GetNextRegs (&vtx_file, regs) != 0)
		    {
		      left = samples_per_regs;
		      AY_SetRegs (regs);
		    }
		  else
		    break;	/*      */
		}

	      //            
	      donow = (need > left) ? left : need;
	      stream = AY_Synth ((Byte*)stream, donow);
	      left -= donow;
	      need -= donow;
	    }			/* end of while(need > 0) */


	  if (stream > &OutBuffer[0])
	    {
	      vtx_ip.add_vis_pcm (vtx_ip.output->written_time (), FMT_S16_LE,
				  2, stream - &OutBuffer[0], &OutBuffer[0]);
	      while (vtx_ip.output->buffer_free () < (stream - &OutBuffer[0])
		     && going && seek_to == -1)
		xmms_usleep (10000);
	      if (going && seek_to == -1)
		vtx_ip.output->write_audio (&OutBuffer[0],
					    stream - &OutBuffer[0]);
	    }
	  else
	    {
	      end = TRUE;
	      vtx_ip.output->buffer_free ();
	      vtx_ip.output->buffer_free ();
	      xmms_usleep (10000);
	    }
	}
      else
	xmms_usleep (10000);
      if (seek_to != -1)
	{
	  vtx_file.pos = seek_to * 50;	/* time in sec * 50 = AY reg data blocks */
	  vtx_ip.output->flush (seek_to * 1000);
	  seek_to = -1;

	}

    }
#endif





  /*     vtx   vtx_stop() */
  pthread_exit (NULL);
}



/*   */
void
vtx_play_file (char *filename)
{
  int rate;
  gchar *buf;

  if (!VTX_ReadHeader (&vtx_file, filename))
    printf ("libvtx: Error read vtx header from %s\n", filename);
  else if (!VTX_ReadData (&vtx_file))
    printf ("libvtx: Error read vtx data from %s\n", filename);
  else
    {
      /* tune AY chip settings */
      AY_SetChip( (strcmp(vtx_file.Chip,"ay") ==0)? _AY:_YM,
        vtx_file.ChipFreq, (stereo_t)vtx_file.Stereo);

      audio_error = FALSE;

      if (vtx_ip.output->open_audio (FMT_S16_NE, OUT_FREQ, 2) == 0)
	{
	  fprintf (stderr, "libvtx: output audio error!\n");
	  audio_error = TRUE;
	  going = FALSE;
	  return;
	}

      AY_Start ();
      end = 0;
      samples_per_regs = OUT_FREQ / vtx_file.PlayerFreq;
      seek_to = -1;

      rate = OUT_FREQ * 2 * 2;
      if ((buf = g_malloc (2 * VTX_NTSTRING_MAX + 40)) != NULL)
	{
	  sprintf (buf, "%s - %s", &vtx_file.Author[0], &vtx_file.Title[0]);

	  vtx_ip.set_info (buf, vtx_file.regdata_size / 14 * 1000 / 50,
			   14 * 50 * 8, OUT_FREQ, 2);
	  g_free (buf);
	}
      going = TRUE;
      pthread_create (&play_thread, NULL, play_loop, NULL);
    }
}



/*   */
void
vtx_stop (void)
{
  if (going)
    {
      going = FALSE;
      pthread_join (play_thread, NULL);
      vtx_ip.output->close_audio ();
      VTX_Free (&vtx_file);
    }
}



/*      */
void
vtx_seek (int time)
{
  //      
  //  time in sec * 50 = register data offset

  //      ?
  if (time * 50 < vtx_file.regdata_size / 14)
    {
      end = 0;
      seek_to = time;

      while (seek_to != -1)
	xmms_usleep (10000);
    }
}



/* Pause or unpause */
void
vtx_pause (short p)
{
  vtx_ip.output->pause (p);
}



/* Get the time, usually returns the output plugins output time */
int
vtx_get_time (void)
{
  if (audio_error)
    return -2;
  if (!going || (end && !vtx_ip.output->buffer_playing ()))
    return -1;
  return vtx_ip.output->output_time ();
}


/* Function to grab the title string */
void
vtx_get_song_info (char *filename, char **title, int *length)
{
  VTXFile tmp;

  (*length) = -1;
  (*title) = NULL;

  if (!VTX_ReadHeader (&tmp, filename))
    fprintf (stderr, "Error Read_VTX_Header %s\n", filename);
  else
    {
      //   
      *length = tmp.regdata_size / 14 * 1000 / 50;

      //  + 
      if ((*title = g_malloc (2 * VTX_NTSTRING_MAX + 40)) != NULL)
	sprintf (*title, "%s - %s", &tmp.Author[0], &tmp.Title[0]);

      VTX_Free (&tmp);
    }
}
